 /*
  * UAE - The Un*x Amiga Emulator
  *
  * lowlevel device glue
  *
  */

#include "sysconfig.h"
#include "sysdeps.h"
#include "options.h"

#include "config.h"

#include "blkdev.h"

#ifdef WIN32

extern struct device_functions devicefunc_win32_aspi;
extern struct device_functions devicefunc_win32_spti;
extern struct device_functions devicefunc_win32_ioctl;
struct device_functions *device_func;

static void install_driver (int flags)
{
    OSVERSIONINFO osVersion;

    device_func = &devicefunc_win32_aspi;
    osVersion.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    if(GetVersionEx(&osVersion)) {
	if(osVersion.dwPlatformId == VER_PLATFORM_WIN32_NT) {
	    if (flags & DEVICE_TYPE_SCSI)
		device_func = &devicefunc_win32_spti;
	    else
		device_func = &devicefunc_win32_ioctl;
	}
    }
    if (currprefs.win32_aspi)
        device_func = &devicefunc_win32_aspi;
    if (device_func == &devicefunc_win32_spti)
        write_log ("Windows NT/2K/XP: Using SPTI CDROM interface\n");
    else if (device_func == &devicefunc_win32_ioctl)
        write_log ("Windows NT/2K/XP: Using IOCTL CDROM interface\n");
    else
	write_log ("Windows 95/98/ME/NT/2K/XP: Using ASPI CDROM interface\n");
}

#endif

void device_func_init (int flags)
{
    static int initialized;

    if (initialized)
	return;
    initialized = 1;
    install_driver (flags);
    device_func->openbus ((flags & DEVICE_TYPE_SCSI) ? 0 : INQ_ROMD);
}

/* pause/unpause CD audio */
void sys_command_pause (int unitnum,int paused)
{
    if (!device_func->pause) {
        uae_u8 cmd[10] = {0x4b,0,0,0,0,0,0,0,paused?0:1,0};
        device_func->exec_out (unitnum, cmd, sizeof (cmd));
	return;
    }
    device_func->pause (unitnum, paused);
}

/* stop CD audio */
void sys_command_stop (int unitnum)
{
    if (!device_func->stop) {
	uae_u8 cmd[6] = {0x4e,0,0,0,0,0};
        device_func->exec_out (unitnum, cmd, sizeof (cmd));
	return;
    }
    device_func->stop (unitnum);
}

/* play CD audio */
int sys_command_play (int unitnum,uae_u32 startmsf, uae_u32 endmsf, int scan)
{
    if (!device_func->play) {
        uae_u8 cmd[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
#if 0
	if (scan) {
	    cmd[0] = 0xba;
	    cmd[1] = scan < 0 ? 0x10 : 0x0;
	    cmd[3] = (uae_u8)(startmsf >> 16);
	    cmd[4] = (uae_u8)(startmsf >> 8);
	    cmd[5] = (uae_u8)(startmsf >> 0);
	    cmd[9] = 0x40;
	} else {
#endif
	    cmd[0] = 0x47;
	    cmd[3] = (uae_u8)(startmsf >> 16);
	    cmd[4] = (uae_u8)(startmsf >> 8);
	    cmd[5] = (uae_u8)(startmsf >> 0);
	    cmd[6] = (uae_u8)(endmsf >> 16);
	    cmd[7] = (uae_u8)(endmsf >> 8);
	    cmd[8] = (uae_u8)(endmsf >> 0);
#if 0
	}
#endif
        return device_func->exec_out (unitnum, cmd, sizeof (cmd)) == 0 ? 0 : 1;
    }
    return device_func->play (unitnum, startmsf, endmsf, scan);
}

/* read qcode */
uae_u8 *sys_command_qcode (int unitnum)
{
    if (!device_func->qcode) {
	uae_u8 cmd[10] = {0x42,2,0x40,1,0,0,0,DEVICE_SCSI_BUFSIZE>>8,DEVICE_SCSI_BUFSIZE&0xff,0};
	return  device_func->exec_in (unitnum, cmd, sizeof (cmd), 0);
    }
    return device_func->qcode (unitnum);
};

/* read table of contents */
uae_u8 *sys_command_toc (int unitnum)
{
    if (!device_func->toc) {
	uae_u8 cmd [10] = { 0x43,0,2,0,0,0,1,DEVICE_SCSI_BUFSIZE>>8,DEVICE_SCSI_BUFSIZE&0xFF,0};
        return device_func->exec_in (unitnum, cmd, sizeof(cmd), 0);
    }
    return device_func->toc (unitnum);
}

/* read one sector */
uae_u8 *sys_command_read (int unitnum, int offset)
{
    if (!device_func->read) {
	uae_u8 cmd[12] = { 0xbe, 0, 0, 0, 0, 0, 0, 0, 1, 0x10, 0, 0 };
	cmd[3] = (uae_u8)(offset >> 16);
	cmd[4] = (uae_u8)(offset >> 8);
	cmd[5] = (uae_u8)(offset >> 0);
        return device_func->exec_in (unitnum, cmd, sizeof (cmd), 0);
    }
    return device_func->read (unitnum, offset);
}
