 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Joystick, mouse and keyboard emulation prototypes and definitions
  *
  * Copyright 1995 Bernd Schmidt
  * Copyright 2001 Toni Wilen
  */

extern void read_joysticks (void);
extern void init_joysticks (void);
extern void close_joysticks (void);
extern int nr_joysticks;

extern void read_mice (void);
extern void init_mice (void);
extern void close_mice (void);
extern int nr_mice;

/* maximum number native input devices supported (single type) */
#define MAX_INPUT_DEVICES 4
/* maximum number of native input device's buttons and axles supported */
#define MAX_INPUT_DEVICE_BUTTONS 32

#define DEFEVENT(A, B, C, D, E) INPUTEVENT_ ## A,
enum inputevents {
INPUTEVENT_ZERO,
#include "inputevents.def"
INPUTEVENT_END
};
#undef DEFEVENT

struct native2amigakey {
    int keycode;
    char asciicode;
    int event;
};

#define JOYBUTTON_1 0 /* fire/left mousebutton */
#define JOYBUTTON_2 1 /* 2nd/right mousebutton */
#define JOYBUTTON_3 2 /* 3rd/middle mousebutton */
#define JOYBUTTON_CD32_PLAY 3
#define JOYBUTTON_CD32_RWD 4
#define JOYBUTTON_CD32_FFW 5
#define JOYBUTTON_CD32_GREEN 6
#define JOYBUTTON_CD32_YELLOW 7
#define JOYBUTTON_CD32_RED 8
#define JOYBUTTON_CD32_BLUE 9

#define INPUTEVENT_JOY1_CD32_FIRST INPUTEVENT_JOY1_CD32_PLAY
#define INPUTEVENT_JOY2_CD32_FIRST INPUTEVENT_JOY2_CD32_PLAY
#define INPUTEVENT_JOY1_CD32_LAST INPUTEVENT_JOY1_CD32_BLUE
#define INPUTEVENT_JOY2_CD32_LAST INPUTEVENT_JOY2_CD32_BLUE

/* event masks */
#define AM_KEY 1 /* this event is only available for keys */
#define AM_JOY_BUT 2 /* joystick buttons only */
#define AM_JOY_AXLE 4 /* joystick axles only */
#define AM_JOY (AM_JOY_BUT|AM_JOY_AXLE)
#define AM_MOUSE_BUT 8 /* mouse buttons only */
#define AM_MOUSE_DIR 16 /* mouse direction only */
#define AM_MOUSE (AM_MOUSE_BUT|AM_MOUSE_DIR)
#define AM_AF 32 /* supports autofire */

/* event flags */
#define ID_FLAG_AUTOREPEAT 1

enum mousestate { mousehack_unknown, mousehack_normal, mousehack_dontcare, mousehack_follow };

extern void handle_cd32_joystick_cia (uae_u8, uae_u8);
extern uae_u8 handle_joystick_buttons (uae_u8);
extern int getbuttonstate (int joy, int button);
extern int getjoystate (int joy);

extern void mousehack_set (enum mousestate);
extern uae_u32 mousehack_helper (void);
extern void mousehack_handle (int sprctl, int sprpos);

extern void setmousebuttonstateall (int mouse, uae_u32 buttonbits, uae_u32 buttonmask);
extern void setjoybuttonstateall (int joy, uae_u32 buttonbits, uae_u32 buttonmask);
extern void setjoybuttonstate (int joy, int button, int state);
extern void setmousebuttonstate (int mouse, int button, int state);
extern void setjoystickstate (int joy, int axle, int state, int max);
extern void setmousestate (int mouse, int x, int y, int isabs);
extern void clearinputstate (void);
extern void translatekeycode (int keycode, char asciicode, int state);
extern void setkeytranslation (struct native2amigakey *trans);

extern uae_u16 potgo_value;
extern uae_u16 POTGOR (void);
extern void POTGO (uae_u16 v);
extern uae_u16 POT0DAT (void);
extern uae_u16 POT1DAT (void);
extern void JOYTEST (uae_u16 v);
extern uae_u16 JOY0DAT (void);
extern uae_u16 JOY1DAT (void);

extern void inputdevice_vsync (void);

extern void write_inputdevice_config (FILE *f);
extern void read_inputdevice_config (char*,char*);
