#include <windows.h>

#include <ddraw.h>

#include <mmsystem.h>
#include <vfw.h>
#include <msacm.h>

#include <stdio.h>

#include "sysconfig.h"
#include "sysdeps.h"
#include "options.h"
#include "custom.h"
#include "picasso96.h"
#include "dxwrap.h"
#include "win32.h"
#include "win32gfx.h"

#include "resource.h"

#include "avioutput.h"

int avioutput_init = 0;

int avioutput_pause = 0;

int avioutput_audio = 0, avioutput_video = 0;

int avioutput_fps = VBLANK_HZ_PAL;

int avioutput_bits = 24;

int avioutput_width = 320, avioutput_height = 256;

int current_width = 320, current_height = 256;

int frame_start = 0; // start frame
int frame_count = 0; // current frame
int frame_end = 0; // end frame (0 = no end, infinite)

int frame_skip = 0;

char avioutput_filename[MAX_PATH] = "C:\\output.avi";

extern char config_filename[256];
extern struct uae_prefs workprefs;



PAVIFILE pfile = NULL; // handle of our AVI file


// audio

PAVISTREAM AVIAudioStream = NULL; // Address of stream interface

HACMSTREAM has = NULL; // stream handle that can be used to perform conversions
ACMSTREAMHEADER ash;

ACMFORMATCHOOSE acmopt;


WAVEFORMATEX wfxSrc; // source audio format
LPWAVEFORMATEX pwfxDst = NULL; // pointer to destination audio format

unsigned char *lpAudio; // pointer to audio data

unsigned int StreamSizeAudio = 0; // audio write position


// video

PAVISTREAM ps = NULL; // Address of stream interface
PAVISTREAM AVIVideoStream = NULL; // compressed stream pointer

AVICOMPRESSOPTIONS videoOptions;
AVICOMPRESSOPTIONS FAR * aOptions[] = {&videoOptions}; // array of pointers to AVICOMPRESSOPTIONS structures

COMPVARS compvars;


LPBITMAPINFOHEADER lpbi = NULL; // bitmap info header + bitmap infomation

unsigned char *lpVideo = NULL; // pointer to video data (bitmap bits)



void AviOutputClearAudioCodec(HWND hwnd)
{
	avioutput_audio = 0;
	
	if(pwfxDst)
	{
		free(pwfxDst);
		pwfxDst = NULL;
	}
	
	CheckDlgButton(hwnd, IDC_AVIOUTPUT_AUDIO, BST_UNCHECKED);
	
	SetWindowText(GetDlgItem(hwnd, IDC_AVIOUTPUT_AUDIO_STATIC), "no codec selected");
	
	if(workprefs.produce_sound < 2)
	{
		EnableWindow(GetDlgItem(hwnd, IDC_AVIOUTPUT_AUDIO), FALSE);
		EnableWindow(GetDlgItem(hwnd, IDC_AVIOUTPUT_AUDIO_STATIC), FALSE);
	}
	else
	{
		EnableWindow(GetDlgItem(hwnd, IDC_AVIOUTPUT_AUDIO), TRUE);
		EnableWindow(GetDlgItem(hwnd, IDC_AVIOUTPUT_AUDIO_STATIC), TRUE);
	}
}



void AviOutputClearVideoCodec(HWND hwnd)
{
	avioutput_video = 0;
	
	if(lpbi)
	{
		free(lpbi);
		lpbi = NULL;
	}
	
	CheckDlgButton(hwnd, IDC_AVIOUTPUT_VIDEO, BST_UNCHECKED);
	
	SetWindowText(GetDlgItem(hwnd, IDC_AVIOUTPUT_VIDEO_STATIC), "no codec selected");
}



UINT CALLBACK acmFilterChooseHookProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if(uMsg == MM_ACM_FORMATCHOOSE)
	{
		switch(wParam)
		{
		case FORMATCHOOSE_FORMATTAG_VERIFY:
			switch(lParam) // remove error prone codecs
			{
			case WAVE_FORMAT_ADPCM: // Microsoft Corporation
				return TRUE;
				
			case WAVE_FORMAT_IMA_ADPCM: // Intel Corporation
				return TRUE;
				
			case WAVE_FORMAT_GSM610: // Microsoft Corporation
				return TRUE;
			}
			break;
		}
	}
	
	return FALSE;
}



LPSTR AVIChooseAudioCodec(HWND hwnd)
{
	DWORD wfxMaxFmtSize;
	
	if(acmMetrics(NULL, ACM_METRIC_MAX_SIZE_FORMAT, &wfxMaxFmtSize) != 0)
	{
		MessageBox(NULL, "acmMetrics FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		return NULL;
	}
	
	if(!(pwfxDst = (LPWAVEFORMATEX) malloc(wfxMaxFmtSize)))
	{
		MessageBox(NULL, "OUT OF MEMORY", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		return NULL;
	}
	
	memset(pwfxDst, 0, wfxMaxFmtSize);
	
	// set the source format
	wfxSrc.wFormatTag = WAVE_FORMAT_PCM;
	wfxSrc.nChannels = workprefs.stereo ? 2 : 1;
	wfxSrc.nSamplesPerSec = workprefs.sound_freq;
	wfxSrc.nBlockAlign = wfxSrc.nChannels * (workprefs.sound_bits / 8);
	wfxSrc.nAvgBytesPerSec = wfxSrc.nBlockAlign * wfxSrc.nSamplesPerSec;
	wfxSrc.wBitsPerSample = workprefs.sound_bits;
	wfxSrc.cbSize = 0;
	
	// set the initial destination format to same as source
	memcpy(pwfxDst, &wfxSrc, sizeof(WAVEFORMATEX));
	
	
	memset(&acmopt, 0, sizeof(ACMFORMATCHOOSE));
	acmopt.cbStruct = sizeof(ACMFORMATCHOOSE);
	acmopt.fdwStyle = ACMFORMATCHOOSE_STYLEF_ENABLEHOOK | ACMFORMATCHOOSE_STYLEF_INITTOWFXSTRUCT;
	acmopt.hwndOwner = hwnd;
	
	acmopt.pwfx = pwfxDst;
	acmopt.cbwfx = wfxMaxFmtSize;
	
	acmopt.pszTitle  = "Choose Audio Codec";
	
	//acmopt.szFormatTag =; // not valid until the format is chosen
	//acmopt.szFormat =; // not valid until the format is chosen
	//acmopt.pszName =; // can use later for config saving loading
	//acmopt.cchName =; // size of pszName, as pszName is non null terminated i think
	
	acmopt.fdwEnum = ACM_FORMATENUMF_INPUT | ACM_FORMATENUMF_NCHANNELS | ACM_FORMATENUMF_NSAMPLESPERSEC;
	//ACM_FORMATENUMF_CONVERT // no, causes WinUAE to become very unstable
	//ACM_FORMATENUMF_WBITSPERSAMPLE // no, because MP3 don't match the bits and so it won't be shown
	//ACM_FORMATENUMF_SUGGEST // no, it only displays MP3 320kbps, closest to src format i suppose
	
	acmopt.pwfxEnum = &wfxSrc;
	
	acmopt.pfnHook = acmFilterChooseHookProc;
	
	switch(acmFormatChoose(&acmopt))
	{
	case MMSYSERR_NOERROR:
		return acmopt.szFormatTag;
		
	case ACMERR_CANCELED:
		//MessageBox(hwnd, "The user chose the Cancel button or the Close command on the System menu to close the dialog box.", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		break;
		
	case ACMERR_NOTPOSSIBLE:
		MessageBox(hwnd, "The buffer identified by the pwfx member of the ACMFORMATCHOOSE structure is too small to contain the selected format.", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		break;
		
	case MMSYSERR_INVALFLAG:
		MessageBox(hwnd, "At least one flag is invalid.", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		break;
		
	case MMSYSERR_INVALHANDLE:
		MessageBox(hwnd, "The specified handle is invalid.", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		break;
		
	case MMSYSERR_INVALPARAM:
		MessageBox(hwnd, "At least one parameter is invalid.", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		break;
		
	case MMSYSERR_NODRIVER:
		MessageBox(hwnd, "A suitable driver is not available to provide valid format selections.", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		break;
		
	default:
		MessageBox(hwnd, "acmFormatChoose FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		break;
	}
	
	return NULL;
}



LPSTR AVIChooseVideoCodec(HWND hwnd)
{
	ICINFO icinfo;
	
	if(!(lpbi = (LPBITMAPINFOHEADER) malloc(sizeof(BITMAPINFOHEADER) + (((avioutput_bits <= 8) ? 1 << avioutput_bits : 0) * sizeof(RGBQUAD)))))
	{
		MessageBox(NULL, "OUT OF MEMORY", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
	}
	
	//memset(lpbi, 0, );
	
	lpbi->biSize = sizeof(BITMAPINFOHEADER);
	
	lpbi->biWidth = avioutput_width;
	lpbi->biHeight = avioutput_height;
	
	lpbi->biPlanes = 1;
	lpbi->biBitCount = (WORD) avioutput_bits;
	lpbi->biCompression = BI_RGB; // uncompressed format
	lpbi->biSizeImage = (lpbi->biWidth * lpbi->biHeight) * (lpbi->biBitCount / 8);
	lpbi->biXPelsPerMeter = 0; // ??
	lpbi->biYPelsPerMeter = 0; // ??
	lpbi->biClrUsed = (lpbi->biBitCount <= 8) ? 1 << lpbi->biBitCount : 0;
	lpbi->biClrImportant = 0;
	
	
	memset(&compvars, 0, sizeof(COMPVARS));
	compvars.cbSize = sizeof(COMPVARS);
	
	// we really should check first to see if the user has a particular compressor installed before we set one
	// we could set one but we will leave it upto the operating system and the set priority levels for the compressors
	
	//compvars.fccHandler = mmioFOURCC('C','V','I','D'); // default to "Cinepak Codec by Radius"
	//compvars.fccHandler = mmioFOURCC('M','R','L','E'); // default to "Microsoft RLE"
	//compvars.fccHandler = mmioFOURCC('D','I','B',' '); // default to "Full Frames (Uncompressed)"
	
	compvars.lQ = 10000; // 10000 is maximum quality setting or ICQUALITY_DEFAULT for default
	compvars.lKey = avioutput_fps; // default to one key frame per second, every (FPS) frames
	
	compvars.dwFlags = ICMF_COMPVARS_VALID;
	
	if(ICCompressorChoose(hwnd, ICMF_CHOOSE_DATARATE | ICMF_CHOOSE_KEYFRAME, lpbi, NULL, &compvars, "Choose Video Codec") == TRUE)
	{
		if(compvars.fccHandler == mmioFOURCC('D','I','B',' '))
			return "Full Frames (Uncompressed)";
		
		// MSDN doesn't specify whether the following structure is to be initialized or not, it works either way
		// But to be on the safe side we will clear the memory
		memset(&icinfo, 0, sizeof(ICINFO));
		//icinfo.dwSize = sizeof(ICINFO); // I think the driver fills this in
		
		if(ICGetInfo(compvars.hic, &icinfo, sizeof(ICINFO)) != 0)
		{
			static char string[128];
			
			if(WideCharToMultiByte(CP_ACP, 0, icinfo.szDescription, -1, string, 128, NULL, NULL) != 0)
				return string;
		}
	}
	
	return NULL;
}



void AVIWriteAudio(uae_u16* sndbuffer, int sndbufsize)
{
    DWORD dwOutputBytes;

    if(acmStreamSize(has, sndbufsize, &dwOutputBytes, ACM_STREAMSIZEF_SOURCE) != 0)    {
    	MessageBox(NULL, "acmStreamSize FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
	AVIUninitialize();
	return;
    }

    if(!(lpAudio = malloc(dwOutputBytes)))
    {
	MessageBox(NULL, "OUT OF MEMORY", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
	AVIUninitialize();
	return;
    }

    ash.cbStruct = sizeof(ACMSTREAMHEADER);
    ash.fdwStatus = 0;
    ash.dwUser = 0;

    // source
    ash.pbSrc = (uae_u8*)sndbuffer;

    ash.cbSrcLength = sndbufsize;
    ash.cbSrcLengthUsed = 0; // This member is not valid until the conversion is complete.

    ash.dwSrcUser = 0;

    // destination
    ash.pbDst = lpAudio;

    ash.cbDstLength = dwOutputBytes;
    ash.cbDstLengthUsed = 0; // This member is not valid until the conversion is complete.

    ash.dwDstUser = 0;

    if(acmStreamPrepareHeader(has, &ash, 0) != 0)
    {
    	MessageBox(NULL, "acmStreamPrepareHeader FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
	AVIUninitialize();
	return;
    }

    if(acmStreamConvert(has, &ash, ACM_STREAMCONVERTF_BLOCKALIGN) != 0)
    {
	MessageBox(NULL, "acmStreamConvert FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
	AVIUninitialize();
	return;
    }

    if(AVIStreamWrite(AVIAudioStream, StreamSizeAudio, pwfxDst->nAvgBytesPerSec / pwfxDst->nBlockAlign, lpAudio, ash.cbDstLengthUsed, 0, NULL, NULL) != 0)
    {
    	MessageBox(NULL, "AVIStreamWrite FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
    	AVIUninitialize();
    	return;
    }

    StreamSizeAudio += pwfxDst->nAvgBytesPerSec / pwfxDst->nBlockAlign;

    acmStreamUnprepareHeader(has, &ash, 0);

    if(lpAudio)
    {
	free(lpAudio);
	lpAudio = NULL;
    }
}


void AVIWriteVideo(void)
{
	HDC hdc;
	HDC hdcMem;
	
	HBITMAP hbitmap;
	HBITMAP hbitmapOld;
	
	
	//if((WIN32GFX_GetWidth() != current_width) || (WIN32GFX_GetHeight() != current_height))
	//{
		current_width = WIN32GFX_GetWidth();
		current_height = WIN32GFX_GetHeight();
	//}
	
	
	if(DirectDraw_GetDC(&hdc, secondary_surface) == DD_OK)
	{
		// create a memory device context compatible with the application's current screen
		hdcMem = CreateCompatibleDC(hdc);
		
		hbitmap = CreateCompatibleBitmap(hdc, avioutput_width, avioutput_height);
		
		hbitmapOld = SelectObject(hdcMem, hbitmap);
		
		
		// probably not a good idea to use slow windows blit functions for this
		// locking the surfaces and copying them by hand would be more efficient perhaps
		
		// draw centered in frame
		
		//BitBlt(hdcMem, currprefs.gfx_xcenter ? (avioutput_width / 2) - (current_width / 2) : 0, currprefs.gfx_xcenter ? (avioutput_height / 2) - (current_height / 2) : 0, current_width, current_height, hdc, 0, 0, SRCCOPY);
		BitBlt(hdcMem, (avioutput_width / 2) - (current_width / 2), (avioutput_height / 2) - (current_height / 2), current_width, current_height, hdc, 0, 0, SRCCOPY);
		
		// normal blit copy
		//BitBlt(hdcMem, 0, 0, current_width, current_height, hdc, 0, 0, SRCCOPY);
		
		
		SelectObject(hdcMem, hbitmapOld);
		
		
		if(GetDIBits(hdc, hbitmap, 0, avioutput_height, lpVideo, (LPBITMAPINFO) lpbi, DIB_RGB_COLORS) == 0)
		{
			MessageBox(NULL, "GetDIBits FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			AVIUninitialize();
			return;
		}
		
		DeleteObject(hbitmap);
		DeleteDC(hdcMem);
		DirectDraw_ReleaseDC(hdc, secondary_surface);
		
		// GetDIBits tends to change this and ruins palettized output
		lpbi->biClrUsed = (avioutput_bits <= 8) ? 1 << avioutput_bits : 0;
		
		if(frame_count == 0)
		{
			if(AVIStreamSetFormat(AVIVideoStream, frame_count, lpbi, lpbi->biSize + (lpbi->biClrUsed * sizeof(RGBQUAD))) != 0)
			{
				MessageBox(NULL, "AVIStreamSetFormat FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
				AVIUninitialize();
				return;
			}
		}
		
		if(AVIStreamWrite(AVIVideoStream,
			frame_count,
			1,
			lpVideo,
			lpbi->biSizeImage,
			0,
			NULL,
			NULL) != 0)
		{
			MessageBox(NULL, "AVIStreamWrite FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			AVIUninitialize();
		}
		
		frame_count++;
		
		if(frame_end)
		{
			if(frame_count >= frame_end)
			{
				AVIUninitialize();
			}
		}
	}
	else
	{
		MessageBox(NULL, "DirectDraw_GetDC FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		AVIUninitialize();
	}
}



void AVIUninitialize(void)
{
	avioutput_init = 0;
	
	if(has)
	{
		acmStreamUnprepareHeader(has, &ash, 0);
		acmStreamClose(has, 0);
	}
	
	if(AVIAudioStream)
		AVIStreamRelease(AVIAudioStream);
	
	if(AVIVideoStream)
		AVIStreamRelease(AVIVideoStream);
	
	if(ps)
		AVIStreamRelease(ps);
	
	if(pfile)
		AVIFileRelease(pfile);
	
	AVIFileExit();
	
	if(lpAudio)
	{
		free(lpAudio);
		lpAudio = NULL;
	}
	
	if(lpVideo)
	{
		free(lpVideo);
		lpVideo = NULL;
	}
	
	if(lpbi)
	{
		free(lpbi);
		lpbi = NULL;
	}
	
	if(pwfxDst)
	{
		free(pwfxDst);
		pwfxDst = NULL;
	}
	
	if(videoOptions.lpParms)
	{
		free(videoOptions.lpParms);
		videoOptions.lpParms = NULL;
	}
}



void AVIInitialize(void)
{
	AVISTREAMINFO avistreaminfo; // Structure containing information about the stream, including the stream type and its sample rate
	
	if(!avioutput_audio && !avioutput_video)
		return;
	
	if(avioutput_init)
	{
		MessageBox(NULL, "AVI already initialized!", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		return;
	}
	
	
	/*
	// VideoForWindowsVersion() function is obsolete
	// GetVersion() or GetVersionEx() should be used instead
	// this is not really necessary considering you need
	// at least win95/98 to run WinUAE in the first place
	
	// check that we are have at least VFW 1.1
	if(HIWORD(VideoForWindowsVersion() < 0x010a))
	{
		MessageBox(NULL, "You are using an older version of Video for Windows and is not compatible!", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		return;
	}
	*/
	
	AVIFileInit();
	
	
	// delete any existing file before writing AVI
	
	SetFileAttributes(avioutput_filename, FILE_ATTRIBUTE_ARCHIVE);
	DeleteFile(avioutput_filename);
	
	// open file for writing
	if(AVIFileOpen(&pfile, avioutput_filename, OF_CREATE | OF_WRITE, NULL) != 0)
	{
		MessageBox(NULL, "AVIFileOpen FAILED\n\nThis can happen if the path or file name was entered incorrectly.", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		AVIUninitialize();
		return;
	}
	
	
	if(avioutput_video)
	{
		if(!(lpVideo = malloc(lpbi->biSizeImage)))
		{
			MessageBox(NULL, "OUT OF MEMORY", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			AVIUninitialize();
			return;
		}
		
		// fill in the header for the video stream
		memset(&avistreaminfo, 0, sizeof(AVISTREAMINFO));
		avistreaminfo.fccType = streamtypeVIDEO; // stream type
		
		// not sure about this, as this is the uncompressed stream, not the compressed stream
		avistreaminfo.fccHandler = compvars.fccHandler;
		
		// incase the amiga changes palette
		if(lpbi->biBitCount < 24)
			avistreaminfo.dwFlags = AVISTREAMINFO_FORMATCHANGES;
		
		//avistreaminfo.dwCaps =; // Capability flags; currently unused
		//avistreaminfo.wPriority =; // Priority of the stream
		//avistreaminfo.wLanguage =; // Language of the stream
		avistreaminfo.dwScale = 1;
		avistreaminfo.dwRate = avioutput_fps; // our playback speed default (PAL 50fps), (NTSC 60fps)
		avistreaminfo.dwStart = 0; // no delay
		avistreaminfo.dwLength = 1; // initial length
		//avistreaminfo.dwInitialFrames =; // audio only
		avistreaminfo.dwSuggestedBufferSize = lpbi->biSizeImage;
		avistreaminfo.dwQuality = -1; // drivers will use the default quality setting
		avistreaminfo.dwSampleSize = 0; // variable video data samples
		
		SetRect(&avistreaminfo.rcFrame, 0, 0, lpbi->biWidth, lpbi->biHeight); // rectangle for stream
		
		//avistreaminfo.dwEditCount =; // Number of times the stream has been edited. The stream handler maintains this count.
		//avistreaminfo.dwFormatChangeCount =; // Number of times the stream format has changed. The stream handler maintains this count.
		strcpy(avistreaminfo.szName, "Videostream"); // description of the stream.
		
		
		// create the stream
		if(AVIFileCreateStream(pfile, &ps, &avistreaminfo) != 0)
		{
			MessageBox(NULL, "AVIFileCreateStream FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			AVIUninitialize();
			return;
		}
		
		
		memset(&videoOptions, 0, sizeof(videoOptions));
		videoOptions.fccType = streamtypeVIDEO;
		videoOptions.fccHandler = compvars.fccHandler;
		videoOptions.dwKeyFrameEvery = compvars.lKey;
		videoOptions.dwQuality = compvars.lQ;
		
		videoOptions.dwBytesPerSecond = compvars.lDataRate * 1024;
		videoOptions.dwFlags = AVICOMPRESSF_VALID | AVICOMPRESSF_KEYFRAMES | AVICOMPRESSF_DATARATE;
		
		videoOptions.dwInterleaveEvery = 0; // no interleaving
		
		videoOptions.cbFormat = sizeof(BITMAPINFOHEADER);
		videoOptions.lpFormat = lpbi;
		
		videoOptions.cbParms = compvars.cbState;
		
		if(!(videoOptions.lpParms = (LPVOID) malloc(compvars.cbState)))
		{
			MessageBox(NULL, "OUT OF MEMORY", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			AVIUninitialize();
			return;
		}
		
		memcpy(videoOptions.lpParms, compvars.lpState, compvars.cbState);
		
		// create a compressed stream from our uncompressed stream and a compression filter
		if(AVIMakeCompressedStream(&AVIVideoStream, ps, &videoOptions, NULL) != AVIERR_OK)
		{
			MessageBox(NULL, "AVIMakeCompressedStream FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			AVIUninitialize();
			return;
		}
		/*
		// we will set the format for the stream on the first frame
		if(AVIStreamSetFormat(AVIVideoStream, 0, lpbi, lpbi->biSize + (lpbi->biClrUsed * sizeof(RGBQUAD))) != AVIERR_OK)
		{
			MessageBox(NULL, "AVIStreamSetFormat FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			return;
		}
		*/
	}
	
	if(avioutput_audio)
	{
		memset(&avistreaminfo, 0, sizeof(AVISTREAMINFO));
		avistreaminfo.fccType = streamtypeAUDIO;
		avistreaminfo.fccHandler = 0; // This member is not used for audio streams.
		avistreaminfo.dwFlags = 0;
		//avistreaminfo.dwCaps =; // Capability flags; currently unused.
		//avistreaminfo.wPriority =;
		//avistreaminfo.wLanguage =;
		avistreaminfo.dwScale = pwfxDst->nBlockAlign;
		avistreaminfo.dwRate = pwfxDst->nAvgBytesPerSec;
		avistreaminfo.dwStart = 0;
		avistreaminfo.dwLength = -1;
		avistreaminfo.dwInitialFrames = 0;
		avistreaminfo.dwSuggestedBufferSize = 0; // Use zero if you do not know the correct buffer size.
		avistreaminfo.dwQuality = -1; // -1 default quality value
		avistreaminfo.dwSampleSize = pwfxDst->nBlockAlign;
		//avistreaminfo.rcFrame; // doesn't apply to audio
		//avistreaminfo.dwEditCount =; // Number of times the stream has been edited. The stream handler maintains this count.
		//avistreaminfo.dwFormatChangeCount =; // Number of times the stream format has changed. The stream handler maintains this count.
		strcpy(avistreaminfo.szName, "Audiostream"); // description of the stream.
		
		
		// create the audio stream
		if(AVIFileCreateStream(pfile, &AVIAudioStream, &avistreaminfo) != 0)
		{
			MessageBox(NULL, "AVIFileCreateStream FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			AVIUninitialize();
			return;
		}
		
		if(AVIStreamSetFormat(AVIAudioStream, 0, pwfxDst, sizeof(WAVEFORMATEX) + pwfxDst->cbSize) != 0)
		{
			MessageBox(NULL, "AVIStreamSetFormat FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			AVIUninitialize();
			return;
		}
		
		if(acmStreamOpen(&has, NULL, &wfxSrc, pwfxDst, NULL, 0, 0, ACM_STREAMOPENF_NONREALTIME) != 0)
		{
			MessageBox(NULL, "acmStreamOpen FAILED", VersionStr, MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
			AVIUninitialize();
			return;
		}
	}
	
	avioutput_init = 1;
}



/*
captures the amiga display surface and saves it to file as a 24bit bitmap
*/
void avioutput_screenshot(void)
{
	// GDI-compatible device context for the surface
	HDC surface_dc = NULL;
	
	// bitmap that is converted to a DIB
	HBITMAP offscreen_bitmap = NULL;
	
	// offscreen DC that we can select offscreen bitmap into
	HDC offscreen_dc = NULL;
	
	// pointer to bitmap bits array
	LPVOID lpvBits = NULL;
	
	// bitmap info
	BITMAPINFO bi;
	
	unsigned int width = WIN32GFX_GetWidth();
	unsigned int height = WIN32GFX_GetHeight();
	
	// Create a GDI-compatible device context for the surface
	if(DirectDraw_GetDC(&surface_dc, secondary_surface) != DD_OK)
	{
		// error
		return;
	}
	
	// need a HBITMAP to convert it to a DIB
	if((offscreen_bitmap = CreateCompatibleBitmap(surface_dc, width, height)) == NULL)
	{
		// error
		return;
	}
	
	// The bitmap is empty, so let's copy the contents of the surface to it.
	// For that we need to select it into a device context.
	if((offscreen_dc = CreateCompatibleDC(surface_dc)) == NULL)
	{
		// error
		return;
	}
	
	
	{
		// select offscreen_bitmap into offscreen_dc
		HGDIOBJ hgdiobj = SelectObject(offscreen_dc, offscreen_bitmap);
		
		// now we can copy the contents of the surface to the offscreen bitmap
		BitBlt(offscreen_dc, 0, 0, width, height, surface_dc, 0, 0, SRCCOPY);
		
		// de-select offscreen_bitmap
		SelectObject(offscreen_dc, hgdiobj);
	}
	
	
	// release surface dc
	DirectDraw_ReleaseDC(surface_dc, secondary_surface);
	surface_dc = NULL;
	
	
	ZeroMemory(&bi.bmiHeader, sizeof(BITMAPINFOHEADER));
	bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi.bmiHeader.biWidth = width;
	bi.bmiHeader.biHeight = height;
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biBitCount = 24;
	bi.bmiHeader.biCompression = BI_RGB;
	bi.bmiHeader.biSizeImage = (bi.bmiHeader.biWidth * bi.bmiHeader.biHeight) * 3;
	bi.bmiHeader.biXPelsPerMeter = 0;
	bi.bmiHeader.biYPelsPerMeter = 0;
	bi.bmiHeader.biClrUsed = 0;
	bi.bmiHeader.biClrImportant = 0;
	
	
	// Reserve memory for bitmap bits
	if(!(lpvBits = malloc(bi.bmiHeader.biSizeImage)))
	{
		// OUT OF MEMORY
		return;
	}
	
	// Have GetDIBits convert offscreen_bitmap to a DIB (device-independent bitmap):
	if(!GetDIBits(offscreen_dc, offscreen_bitmap, 0, bi.bmiHeader.biHeight, lpvBits, &bi, DIB_RGB_COLORS))
	{
		// GetDIBits FAILED
		return;
	}
	
	if(offscreen_dc)
		DeleteDC(offscreen_dc);
	
	if(offscreen_bitmap)
		DeleteObject(offscreen_bitmap);
	
	{
		int i = 0;
		FILE *fp;
		char filename[256];
		strcpy(filename, config_filename);
		strcat(filename, ".bmp");
		
		while(++i < 1000) // limit 999 screenshots
		{
			if((fp = fopen(filename, "r")) == NULL) // does file not exist?
			{
				BITMAPFILEHEADER bfh;
				
				if((fp = fopen(filename, "wb")) == NULL)
				{
					// error
					return;
				}
				
				// write the file header, bitmap information and pixel data
				bfh.bfType = 19778;
				bfh.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + bi.bmiHeader.biSizeImage;
				bfh.bfReserved1 = 0;
				bfh.bfReserved2 = 0;
				bfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
				
				if(fwrite(&bfh, 1, sizeof(BITMAPFILEHEADER), fp) < sizeof(BITMAPFILEHEADER))
				{
					// failed to write file header
					
					fclose(fp);
					return;
				};
				
				if(fwrite(&bi, 1, sizeof(BITMAPINFOHEADER), fp) < sizeof(BITMAPINFOHEADER))
				{
					// failed to write bitmap header
					
					fclose(fp);
					return;
				};
				
				if(fwrite(lpvBits, 1, bi.bmiHeader.biSizeImage, fp) < bi.bmiHeader.biSizeImage)
				{
					// failed to write the bitmap
					
					fclose(fp);
					return;
				};
				
				fclose(fp);
				break;
			}
			else
			{
				char tempstring[256];
				char number[256];
				int x = 0;
				int y = 0;
				fclose(fp);
				
				filename[strlen(filename) - 4] = '\0'; // trim filename extension
				
				x = strlen(filename) - 1;
				
				while((filename[x] >= '0') && (filename[x] <= '9'))
				{
					x--;
					if(x < 1)
						break;
				}
				
				x++;
				
				while(filename[x + y] != '\0')
				{
					number[y] = filename[x + y];
					y++;
				}
				
				number[y] = '\0';
				
				filename[x] = '\0';
				
				strcat(filename, itoa(atoi(number) + 1, tempstring, 10));
				
				strcat(filename, ".bmp");
			}
		}
	}
	
	if(lpvBits)
		free(lpvBits);
	
	return;
}

void frame_drawn (void)
{
    if(avioutput_init && avioutput_video)
	AVIWriteVideo();
}


