 /*
  * UAE - The Un*x Amiga Emulator
  *
  * WIN32 CDROM/HD low level access code (SPTI)
  *
  * Copyright 2002 Toni Wilen
  *
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include "memory.h"

#include "threaddep/thread.h"
#include "blkdev.h"

#include <stddef.h>

#include <windows.h>
#include <devioctl.h>
#include <ntddscsi.h>

typedef struct {
  SCSI_PASS_THROUGH_DIRECT spt;
  ULONG Filler;
  UCHAR SenseBuf[32];
} SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER;

static int unitcnt = 0;

static uae_sem_t scgp_sem;
static char driveletters[MAX_TOTAL_DEVICES * 8];
static char *drvletter[MAX_TOTAL_DEVICES];
static HANDLE handles[MAX_TOTAL_DEVICES];
static uae_u8 *scsibuf;

static int doscsi (HANDLE *h, SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER *swb)
{
    DWORD status, returned;

#if 0
    int i;
    device_debug("SCSI, H=%08.8X: ",h);
    for (i = 0; i < swb->spt.CdbLength; i++) {
	device_debug ("%s%02.2X", i > 0 ? "." : "", swb->spt.Cdb[i]);
    }
    device_debug("\n");
#endif
    status = DeviceIoControl (h, IOCTL_SCSI_PASS_THROUGH_DIRECT,
	swb, sizeof (SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER),
	swb, sizeof (SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER),
	&returned, NULL);
    if (!status) {
	int lasterror = GetLastError();
	device_debug ("scsi error: %d, lasterror=%d\n", swb->spt.ScsiStatus, lasterror);
    }
#if 0
    if (swb->spt.SenseInfoLength>0 && swb->spt.SenseInfoLength < 32) {
	device_debug("SENSE: ");
	for (i = 0; i < swb->spt.SenseInfoLength; i++) {
	    device_debug ("%s%02.2X", i > 0 ? "." : "", swb->SenseBuf[i]);
	}
        device_debug("\n");
	return 0;
    }
#endif
    return status;
}

static int execscsicmd (int unitnum, uae_u8 *data, int len, uae_u8 *inbuf, int inlen)
{
    SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER swb;
    DWORD status;

    uae_sem_wait (&scgp_sem);
    memset (&swb, 0, sizeof (SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER));
    swb.spt.Length = sizeof (SCSI_PASS_THROUGH);
    swb.spt.CdbLength = len;
    if (inbuf) {
	swb.spt.DataIn = SCSI_IOCTL_DATA_IN;
	swb.spt.DataTransferLength = inlen;
	swb.spt.DataBuffer = inbuf;
	memset (inbuf, 0, inlen);
    } else {
	swb.spt.DataIn = SCSI_IOCTL_DATA_OUT;
    }
    swb.spt.TimeOutValue = 80 * 60;
    swb.spt.SenseInfoOffset = offsetof(SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER, SenseBuf);
    swb.spt.SenseInfoLength = 32;
    memcpy (swb.spt.Cdb, data, len);
    status = doscsi (handles[unitnum], &swb);
    uae_sem_post (&scgp_sem);
    if (!status) return -1;
    return swb.spt.DataTransferLength;
}

static int execscsicmd_direct (int unitnum, uaecptr acmd)
{
    SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER swb;
    DWORD status;
    int sactual = 0, i;
    uaecptr scsi_data = get_long (acmd + 0);
    uae_u32 scsi_len = get_long (acmd + 4);
    uaecptr scsi_cmd = get_long (acmd + 12);
    uae_u16 scsi_cmd_len = get_word (acmd + 16);
    uae_u8 scsi_flags = get_byte (acmd + 20);
    uaecptr scsi_sense = get_long (acmd + 22);
    uae_u16 scsi_sense_len = get_word (acmd + 26);
    int io_error = 0;
    addrbank *bank_data = &get_mem_bank (scsi_data);

    memset (&swb, 0, sizeof (SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER));
    swb.spt.Length = sizeof (SCSI_PASS_THROUGH);
    swb.spt.SenseInfoOffset = offsetof(SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER, SenseBuf);

    /* do transfer directly to and from Amiga memory */
    if (!bank_data || !bank_data->check (scsi_data, scsi_len))
        return -5; /* IOERR_BADADDRESS */

    uae_sem_wait (&scgp_sem);

    /* the Amiga does not tell us how long the timeout shall be, so make it _very_ long (specified in seconds) */
    swb.spt.TimeOutValue = 80 * 60;
    swb.spt.DataBuffer = bank_data->xlateaddr (scsi_data);
    swb.spt.DataTransferLength = scsi_len;
    swb.spt.DataIn = (scsi_flags & 1) ? SCSI_IOCTL_DATA_IN : SCSI_IOCTL_DATA_OUT;
    swb.spt.CdbLength = (UCHAR)scsi_cmd_len;
    for (i = 0; i < scsi_cmd_len; i++)
	swb.spt.Cdb[i] = get_byte (scsi_cmd + i);
    swb.spt.SenseInfoLength  = (scsi_flags & 4) ? 4 : /* SCSIF_OLDAUTOSENSE */
        (scsi_flags & 2) ? scsi_sense_len : /* SCSIF_AUTOSENSE */
        32;
    status = doscsi (handles[unitnum], &swb);

    put_word (acmd + 18, status == 0 ? 0 : scsi_cmd_len); /* fake scsi_CmdActual */
    put_byte (acmd + 21, swb.spt.ScsiStatus); /* scsi_Status */
    if (swb.spt.ScsiStatus) {
        io_error = 45; /* HFERR_BadStatus */
        /* copy sense? */
        for (sactual = 0; scsi_sense && sactual < scsi_sense_len && sactual < swb.spt.SenseInfoLength; sactual++)
            put_byte (scsi_sense + sactual, swb.SenseBuf[sactual]);
        put_long (acmd + 8, 0); /* scsi_Actual */
    } else {
        int i;
        for (i = 0; i < scsi_sense_len; i++)
            put_byte (scsi_sense + i, 0);
        sactual = 0;
        if (status == 0) {
	    io_error = 20; /* io_Error, but not specified */
	    put_long (acmd + 8, 0); /* scsi_Actual */
        } else {
	    io_error = 0;
            put_long (acmd + 8, scsi_len); /* scsi_Actual */
        }
    }
    put_word (acmd + 28, sactual);

    uae_sem_post (&scgp_sem);
    return io_error;
}

static uae_u8 *execscsicmd_out (int unitnum, uae_u8 *data, int len)
{
    int v = execscsicmd (unitnum, data, len, 0, 0);
    if (v < 0)
	return 0;
    return data;
}

static uae_u8 *execscsicmd_in (int unitnum, uae_u8 *data, int len, int *outlen)
{
    int v = execscsicmd (unitnum, data, len, scsibuf, DEVICE_SCSI_BUFSIZE);
    if (v < 0)
	return 0;
    if (v == 0)
	return 0;
    if (outlen)
	*outlen = v;
    return scsibuf;
}

static int adddrive (char *dev, int *offset, int drive)
{
    if (unitcnt >= MAX_TOTAL_DEVICES)
	return 0;
    sprintf (dev + (*offset), "\\\\.\\%c:", drive);
    drvletter[unitcnt] = dev + (*offset);
    *offset += 8;
    write_log ("CDROM: using drive %c:\n", drive);
    unitcnt++;
    return 1;
}

static total_devices;

static int open_scsi_bus (int flags)
{
    int dwDriveMask;
    int drive, firstdrive;
    char *dev, tmp[10], tmp2[10];
    int i, offset;

    offset = 0;
    total_devices = 0;
    firstdrive = 0;
    uae_sem_init (&scgp_sem, 0, 1);
    for (i = 0; i < MAX_TOTAL_DEVICES; i++)
	handles[i] = INVALID_HANDLE_VALUE;
    if (!scsibuf)
	scsibuf = VirtualAlloc (NULL, DEVICE_SCSI_BUFSIZE, MEM_COMMIT, PAGE_READWRITE);
    dev = driveletters;
    dwDriveMask = GetLogicalDrives();
    dwDriveMask >>= 2; // Skip A and B drives...
    for( drive = 'C'; drive <= 'Z'; drive++) {
	if (dwDriveMask & 1) {
	    sprintf( tmp, "%c:\\", drive );
	    sprintf( tmp2, "%c:\\.", drive );
	    if (GetDriveType (tmp) == DRIVE_CDROM) {
		if (!firstdrive) firstdrive = drive;
		if (adddrive (dev, &offset, drive))
		    total_devices++;
	    }
	}
	dwDriveMask >>= 1;
    }
    return total_devices;
}

static void close_scsi_bus (void)
{
    VirtualFree (scsibuf, 0, MEM_RELEASE);
    scsibuf = 0;
}

int open_scsi_device (int unitnum)
{
    HANDLE h;
    char *dev;

    dev = drvletter[unitnum];
    if (!dev) return 0;
    h = CreateFile(dev,GENERIC_READ|GENERIC_WRITE,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
    if (h == INVALID_HANDLE_VALUE) {
        device_debug ("failed to open cd unit %d (%s)\n", unitnum, dev);
    } else {
        device_debug ("cd unit %d opened (%s)\n", unitnum, dev);
    }
    handles[unitnum] = h;
    return handles[unitnum] == INVALID_HANDLE_VALUE ? 0 : 1;
}

static void close_scsi_device (int unitnum)
{
    device_debug ("cd unit %d closed\n", unitnum);
    if (handles[unitnum] != INVALID_HANDLE_VALUE)
	CloseHandle (handles[unitnum]);
    handles[unitnum] = INVALID_HANDLE_VALUE;
}

static int mediacheck (int unitnum)
{
    uae_u8 cmd [10] = { 0x43,0,2,0,0,0,1,DEVICE_SCSI_BUFSIZE>>8,DEVICE_SCSI_BUFSIZE&0xFF,0};
    if (handles[unitnum] == INVALID_HANDLE_VALUE)
	return 0;
    return execscsicmd_in (unitnum, cmd, sizeof(cmd), 0) ? 1 : 0;
}

static struct device_info *info_device (int unitnum, struct device_info *di)
{
    if (unitnum >= MAX_TOTAL_DEVICES || !drvletter[unitnum])
	return 0;
    sprintf(di->label,"Drive %s", drvletter[unitnum] + 4);
    di->bus = 0;
    di->target = unitnum;
    di->lun = 0;
    di->media_inserted = mediacheck (unitnum);
    di->write_protected = 1;
    di->sectorsize = 2048;
    di->type = INQ_ROMD;;
    return di;
}


struct device_functions devicefunc_win32_spti = {
    open_scsi_bus, close_scsi_bus, open_scsi_device, close_scsi_device, info_device,
    execscsicmd_out, execscsicmd_in, execscsicmd_direct,
    0, 0, 0, 0, 0
};
